/* Copyright (c) 1999,2000 by Howard R. Cole.  All Rights Reserved. */

#import "PGPserver.h"
#import "PGPerrs.h"
@class NSColor;

static PGPError encodeEvent(PGPContextRef, PGPEvent *, PGPUserValue);
static PGPError decodeEvent(PGPContextRef, PGPEvent *, PGPUserValue);

@implementation PGPserver

- init
{
    id fileMan = [NSFileManager defaultManager];
    BOOL isDir;
    NSUserDefaults *pgpDefaulter = [NSUserDefaults standardUserDefaults];
    PGPFileSpecRef defaultsFileSpec;
    NSString *defaultsFile;
    PGPError err;

    [super init];
    
    doSignature = NO;
    detachSignature = NO;
    clearSign = NO;
    doArmor = YES;
    doTextTranslation = NO;
    doConventionalOnly = NO;
    useUntrusted = YES;
    encryptToSelf = NO;

    entropyBucket = NULL;
    entropyLength = 0;

    /* Get default values for PGP directory */
    pgpHomeDir = [[pgpDefaulter stringForKey:@"PGPHomeDirectory"] retain];
    if(pgpHomeDir==nil)
        pgpHomeDir = [[NSHomeDirectory() stringByAppendingPathComponent:@".pgp"] retain];

    secRingFile = [[pgpDefaulter stringForKey:@"PGPSecretKeyRing"] retain];
    if(secRingFile == nil)
	secRingFile = [[pgpHomeDir stringByAppendingPathComponent:@"secring.skr"] retain];

    pubRingFile = [[pgpDefaulter stringForKey:@"PGPPublicKeyRing"] retain];
    if(pubRingFile == nil) 
	pubRingFile = [[pgpHomeDir stringByAppendingPathComponent:@"pubring.pkr"] retain];
    
    /* Check whether .pgp exists */
    if(![fileMan fileExistsAtPath:pgpHomeDir isDirectory:&isDir]) {
        fprintf(stderr,"PGPserver: %s",noPGPdir);
        return nil;
    }

    /* Check whether .pgp is really a directory */
    if(!isDir) {
	fprintf(stderr,"PGPserver: %s",pgpNotDir);
	return nil;
    }


    /* Initialize the SDK */
    err = PGPsdkInit();
    if((err != 0)) {
       fprintf(stderr,"PGP error initializing SDK\n");
       return nil;
    }

    err = PGPNewContext(kPGPsdkAPIVersion, &context);
    if((err != 0)) {
       fprintf(stderr,"PGP error creating PGP context\n");
       return nil;
    }

    /* Create a fileSpec for defaults file */
    defaultsFile = [pgpHomeDir stringByAppendingPathComponent:@"PGPsdkPreferences"];
    err = PGPNewFileSpecFromFullPath(context, [defaultsFile cString], &defaultsFileSpec);
    if(err != 0) {
        fprintf(stderr,"PGP Error %d creating file spec ref for defaults\n",err);
        return nil;
    }

    err = PGPsdkLoadPrefs(context,defaultsFileSpec);
    if(err != 0) {
        fprintf(stderr,"Can't load PGP defaults from file %s\n",[defaultsFile cString]);
    }

    PGPFreeFileSpec(defaultsFileSpec);

    keyRingSet = NULL;
    if([self openKeyringFiles:NO]==nil)
      if([self openKeyringFiles:YES]==nil) return nil;

    return self;
}

- (void)dealloc
{
    [pgpHomeDir release];
    [secRingFile release];
    [pubRingFile release];

    PGPFreeKeySet(keyRingSet);
    PGPFreeContext(context);
    PGPsdkCleanup();

    return;
}

- openKeyringFiles:(BOOL)writeableFlag
{
    PGPFileSpecRef secRingFileSpec, pubRingFileSpec;
    char errorText[256];
    PGPError err;

    /* Close out the old keyrings */
    if(keyRingSet) {
        err = PGPFreeKeySet(keyRingSet);
        if(err != 0) {
            keyRingSet = NULL;
            PGPGetErrorString(err, 256, errorText);
            fprintf(stderr,"PGPError %d freeing keyRingSet: %s\n",err,errorText);
            return nil;
        }
    }

    err = PGPNewFileSpecFromFullPath(context, [secRingFile cString], &secRingFileSpec);
    if(err != 0) {
        PGPGetErrorString(err, 256, errorText);
	fprintf(stderr,"PGP error %d getting Secret Ring File Spec: %s\n",err,errorText);
	return nil;
    }
    
    err = PGPNewFileSpecFromFullPath(context, [pubRingFile cString], &pubRingFileSpec);
    if(err != 0) {
	fprintf(stderr,"PGP error getting Public Ring File Spec\n");
	return nil;
    }

    if(writeableFlag==YES)
        err = PGPOpenKeyRingPair(context, kPGPKeyRingOpenFlags_Mutable | kPGPKeyRingOpenFlags_Create,
                                 pubRingFileSpec, secRingFileSpec, &keyRingSet);
    else
        err = PGPOpenKeyRingPair(context, kPGPKeyRingOpenFlags_None,
                                 pubRingFileSpec, secRingFileSpec, &keyRingSet);
        
    if(err != 0) {
        PGPGetErrorString(err, 256, errorText);
	fprintf(stderr,"PGP error %d opening Key Ring pair: %s\n",err,errorText);
        keyRingSet = NULL;
	return nil;
    }

    err = PGPFreeFileSpec(secRingFileSpec);
    if(err != 0)
	fprintf(stderr,"PGP error freeing Secret Ring File Spec\n");

    err = PGPFreeFileSpec(pubRingFileSpec);
    if(err != 0)
	fprintf(stderr,"PGP error freeing Public Ring File Spec\n");

    return self;
}


- pgpPublicRingFile:(NSString *)newPubRing writeable:(BOOL)writeFlag
{
    [newPubRing retain];
    if(pubRingFile) [pubRingFile release];
    pubRingFile = newPubRing;
    [self openKeyringFiles:writeFlag];
    return self;
}


- pgpSecretRingFile:(NSString *)newSecRing writeable:(BOOL)writeFlag
{
    [newSecRing retain];
    if(secRingFile) [secRingFile release];
    secRingFile = newSecRing;
    [self openKeyringFiles:writeFlag];
    return self;
}

- pgpHomeDir:(NSString *)newHome writeable:(BOOL)writeFlag
{
    [newHome retain];
    if(pgpHomeDir) [pgpHomeDir release];
    pgpHomeDir = newHome;
    [self openKeyringFiles:writeFlag];
    return self;
}




- doSignature:(BOOL)useSig
{
    doSignature = useSig;
    return self;
}

- clearSign:(BOOL)flag
{
    clearSign = flag;
    detachSignature = !flag;
    return self;
}

- detachSignature:(BOOL)detach
{
    detachSignature = detach;
    clearSign = !detach;
    return self;
}


- encryptToSelf:(BOOL)flag
{
    encryptToSelf = flag;
    return self;
}


- doConventionalOnly:(BOOL)flag
{
    doConventionalOnly = flag;
    return self;
}


- doArmor:(BOOL)useArmor
{
    doArmor = useArmor;
    return self;
}

- doTextTranslation:(BOOL)useText
{
    doTextTranslation = useText;
    return self;
}

- (BOOL)doConventionalOnly
{
    return doConventionalOnly;
}

- (BOOL)clearSign
{
    return clearSign;
}


- useUntrustedKeys:(BOOL)untrustedFlag
{
    useUntrusted = untrustedFlag;
    return self;
}


/***************************************  ENCRYPTION ROUTINES ***********************************/

- (NSData *)encryptData:(NSData *)plainText 
                      withKeys:(NSMutableArray *)userIDArray
                    passPhrase:(NSString *)passPhraseString
                         error:(int *)errorCode
{
    PGPOptionListRef optionList;
    PGPError err;
    int iuser;
    PGPFilterRef recipFilter;
    PGPKeySetRef tmpKeySet,recipKeySet;
    PGPUInt32 numKeys;
    PGPSize cipherLength,maxCipherLength;
    char *cipherBuffer;
    NSData *cipherData;
    char errorText[256];
    static PGPError encodeEvent(PGPContextRef, PGPEvent *, PGPUserValue);

    /* Initialize error code result */
    *errorCode = 0;
    maxCipherLength = 1000000000;

    /* Make a new Option List */
    err = PGPNewOptionList(context, &optionList);
    if(err != 0) {
        fprintf(stderr,"PGP Error %d creating encryption option list\n",err);
        *errorCode = err;
        return nil;
    }

    if(doArmor) {
        err = PGPAppendOptionList(optionList,
                                  PGPOArmorOutput(context, TRUE),
                                  PGPOLastOption(context));

        if(err != 0) {
            fprintf(stderr,"PGP Error %d appending ascii armor option to option list\n",err);
            *errorCode = err;
            return nil;
        }
    }

    /* Append necessary options */
    err = PGPAppendOptionList(optionList,
                              PGPOEventHandler(context, encodeEvent, self),
                              PGPOInputBuffer(context, (char *)[plainText bytes], (int)[plainText length]),
                              PGPOAllocatedOutputBuffer(context, (void **)&cipherBuffer,
                                                        maxCipherLength,(PGPSize *)&cipherLength),
                              PGPOLastOption(context));

    if(err != 0) {
        fprintf(stderr,"PGP Error %d appending triplet options to option list\n",err);
        *errorCode = err;
        return nil;
    }

    if(doConventionalOnly==YES) {
        err = PGPAppendOptionList(optionList,
                                  PGPOConventionalEncrypt(context,
                                                          PGPOPassphrase(context, [passPhraseString cString]),
                                                          PGPOLastOption(context)),
                                  PGPOLastOption(context));

        if(err != 0) {
            fprintf(stderr,"PGP Error %d appending conventional encryption to option list\n",err);
            *errorCode = err;
            return nil;
        }
    }
    else {
        
        /* Make a new keyset for recipient userIDs */
        err = PGPNewKeySet(context, &recipKeySet);
        if(err != 0) {
            fprintf(stderr,"PGP Error %d creating a recipient keyset\n",err);
            *errorCode = err;
            return nil;
        }

        if(encryptToSelf)
            [userIDArray addObject:NSUserName()];

        /* For each userID, add their encryption key to keyset */
        for(iuser=0; iuser<[userIDArray count]; iuser++) {
            const char *recipID;

            recipID = (char *)[[userIDArray objectAtIndex:iuser] cString];

            err = PGPNewUserIDStringFilter(context, recipID, kPGPMatchSubString, &recipFilter);
            if((err != 0)) {
                fprintf(stderr,"PGP error %d creating userID filter\n",err);
                *errorCode = err;
                return nil;
            }

            /* Does this return an error if recipient isn't on keyring? */
            err = PGPFilterKeySet(keyRingSet, recipFilter, &tmpKeySet);
            if((err != 0)) {
                PGPGetErrorString(err, 256, errorText);
                fprintf(stderr,"PGP error %d Filtering KeySet: %s\n",err,errorText);
            }

            /* Check whether you found the key for this user */
            PGPCountKeys(tmpKeySet, &numKeys);
            if(numKeys > 0) {
                err = PGPAddKeys(tmpKeySet, recipKeySet);
                if(err != 0) {
                    fprintf(stderr,"PGP Error %d adding key for user %s to keyset\n",err,recipID);
                }
            }

            err = PGPFreeFilter(recipFilter);
            if(err != 0) {
                fprintf(stderr,"PGP Error %d freeing key filter for user %s\n",err,recipID);
            }

            err = PGPFreeKeySet(tmpKeySet);
            if(err != 0) {
                fprintf(stderr,"PGP Error %d freeing temporary keyset for user %s\n",err,recipID);
            }
        }

        err = PGPAppendOptionList(optionList,
                                  PGPOEncryptToKeySet(context, recipKeySet),
                                  PGPOLastOption(context));

        if(err != 0) {
            fprintf(stderr,"PGP Error %d appending Key Set to option list\n",err);
            *errorCode = err;
            return nil;
        }

        err = PGPFreeKeySet(recipKeySet);
        if(err != 0) {
            fprintf(stderr,"PGP Error %d freeing recipient keyset\n",err);
        }

        if(doSignature == YES) {
            PGPKeyRef sigKey;

            err = PGPGetDefaultPrivateKey(keyRingSet, &sigKey);
            if(err == 0) {
                err = PGPAppendOptionList(optionList,
                                          PGPOSignWithKey(context,
                                                          sigKey,
                                                          PGPOPassphrase(context, [passPhraseString cString]),
                                                          PGPOLastOption(context)),
                                          PGPOLastOption(context));

                if(err != 0) {
                    fprintf(stderr,"PGP Error %d appending signature option to option list\n",err);
                    *errorCode = err;
                    doSignature = NO;
                    clearSign = NO;
                }
            }
            else {
                PGPGetErrorString(err, 256, errorText);
                fprintf(stderr,"PGP error %d getting default signing key: %s\n",err,errorText);
                *errorCode = err;
                doSignature = NO;
                clearSign = NO;
            }
        }
    }

    err = PGPEncode(context,optionList,PGPOLastOption(context));
    if(err != 0) {
        fprintf(stderr,"PGP Error %d on encrypt\n",err);
        *errorCode = err;
        return nil;
    }

    err = PGPFreeOptionList(optionList);
    if(err != 0) {
        fprintf(stderr,"PGP Error %d freeing encryption option list\n",err);
    }

    cipherData = [NSData dataWithBytes:cipherBuffer length:cipherLength];
    PGPFreeData(cipherBuffer);

    return cipherData;

}

- encryptFile:(NSString *)plainFile toFile:(NSString *)cipherFile 
                                  withKeys:(NSMutableArray *)userIDArray
                                passPhrase:(NSString *)passPhraseString
                                     error:(int *)errorCode
{
    PGPOptionListRef optionList;
    PGPFileSpecRef plainFileSpec,cipherFileSpec;
    PGPError err;
    int iuser;
    PGPFilterRef recipFilter;
    PGPKeySetRef tmpKeySet,recipKeySet;
    PGPUInt32 numKeys;
    char errorText[256];
    static PGPError encodeEvent(PGPContextRef, PGPEvent *, PGPUserValue);

    /* Initialize error code result */
    *errorCode = 0;

    /* Get the plainText FileSpecRef */
    err = PGPNewFileSpecFromFullPath(context, [plainFile cString], &plainFileSpec);
    if(err != 0) {
        fprintf(stderr,"PGP Error %d creating plain text file spec ref list\n",err);
        *errorCode = err;
        return nil;
    }

    /* Get the cipherText FileSpecRef */
    err = PGPNewFileSpecFromFullPath(context, [cipherFile cString], &cipherFileSpec);
    if(err != 0) {
        fprintf(stderr,"PGP Error %d creating cipher text file spec ref list\n",err);
        *errorCode = err;
        return nil;
    }

    /* Make a new Option List */
    err = PGPNewOptionList(context, &optionList);
    if(err != 0) {
        fprintf(stderr,"PGP Error %d creating encryption option list\n",err);
        *errorCode = err;
        return nil;
    }

    if(doArmor) {
        err = PGPAppendOptionList(optionList,
                                  PGPOArmorOutput(context, TRUE),
                                  PGPOLastOption(context));

        if(err != 0) {
            fprintf(stderr,"PGP Error %d appending ascii armor option to option list\n",err);
            *errorCode = err;
            return nil;
        }
    }

    /* Append necessary options */
    err = PGPAppendOptionList(optionList,
                              PGPOEventHandler(context, encodeEvent, self),
                              PGPOInputFile(context, plainFileSpec),
                              PGPOOutputFile(context, cipherFileSpec),
                              PGPOLastOption(context));

    if(err != 0) {
        fprintf(stderr,"PGP Error %d appending triplet options to option list\n",err);
        *errorCode = err;
        return nil;
    }

    if(doConventionalOnly==YES) {
        err = PGPAppendOptionList(optionList,
                                  PGPOConventionalEncrypt(context,
                                                          PGPOPassphrase(context, [passPhraseString cString]),
                                                          PGPOLastOption(context)),
                                  PGPOLastOption(context));

        if(err != 0) {
            fprintf(stderr,"PGP Error %d appending conventional encryption to option list\n",err);
            *errorCode = err;
            return nil;
        }
    }
    else {

        /* Make a new keyset for recipient userIDs */
        err = PGPNewKeySet(context, &recipKeySet);
        if(err != 0) {
            fprintf(stderr,"PGP Error %d creating a recipient keyset\n",err);
            return nil;
        }

        if(encryptToSelf)
            [userIDArray addObject:NSUserName()];

        /* For each userID, add their encryption key to keyset */
        for(iuser=0; iuser<[userIDArray count]; iuser++) {
            char *recipID;

            recipID = (char *)[[userIDArray objectAtIndex:iuser] cString];

            err = PGPNewUserIDStringFilter(context, recipID, kPGPMatchSubString, &recipFilter);
            if((err != 0)) {
                fprintf(stderr,"PGP error %d creating userID filter\n",err);
                *errorCode = err;
                return nil;
            }

            err = PGPFilterKeySet(keyRingSet, recipFilter, &tmpKeySet);
            if((err != 0)) {
                fprintf(stderr,"PGP error Filtering KeySet\n");
                *errorCode = err;
                return nil;
            }

            /* Check whether you found the key for this user */
            PGPCountKeys(tmpKeySet, &numKeys);
            if(numKeys > 0) {
                err = PGPAddKeys(tmpKeySet, recipKeySet);
                if(err != 0) {
                    fprintf(stderr,"PGP Error %d adding key for user %s to keyset\n",err,recipID);
                }
            }

            err = PGPFreeFilter(recipFilter);
            if(err != 0) {
                fprintf(stderr,"PGP Error %d freeing key filter for user %s\n",err,recipID);
            }

            err = PGPFreeKeySet(tmpKeySet);
            if(err != 0) {
                fprintf(stderr,"PGP Error %d freeing temporary keyset for user %s\n",err,recipID);
            }

        }

        err = PGPAppendOptionList(optionList,
                                  PGPOEncryptToKeySet(context, recipKeySet),
                                  PGPOLastOption(context));

        if(err != 0) {
            fprintf(stderr,"PGP Error %d appending Key Set to option list\n",err);
            *errorCode = err;
            return nil;
        }

        err = PGPFreeKeySet(recipKeySet);
        if(err != 0) {
            fprintf(stderr,"PGP Error %d freeing recipient keyset\n",err);
        }

        if(doSignature == YES) {
            PGPKeyRef sigKey;

            err = PGPGetDefaultPrivateKey(keyRingSet, &sigKey);
            if(err == 0) {
                err = PGPAppendOptionList(optionList,
                                          PGPOSignWithKey(context,
                                                          sigKey,
                                                          PGPOPassphrase(context, [passPhraseString cString]),
                                                          PGPOLastOption(context)),
                                          PGPOLastOption(context));

                if(err != 0) {
                    fprintf(stderr,"PGP Error %d appending signature option to option list\n",err);
                    *errorCode = err;
                    doSignature = NO;
                    clearSign = NO;
                }
            }
            else {
                PGPGetErrorString(err, 256, errorText);
                fprintf(stderr,"PGP error %d getting default signing key: %s\n",err,errorText);
                *errorCode = err;
            }
        }
    }

    err = PGPEncode(context,optionList,PGPOLastOption(context));
    if(err != 0) {
        fprintf(stderr,"PGP Error %d on encrypt\n",err);
        *errorCode = err;
        return nil;
    }

    err = PGPFreeFileSpec(plainFileSpec);
    if(err != 0) {
        fprintf(stderr,"PGP Error %d freeing plain file spec\n",err);
    }


    err = PGPFreeFileSpec(cipherFileSpec);
    if(err != 0) {
        fprintf(stderr,"PGP Error %d freeing cipher file spec\n",err);
    }


    err = PGPFreeOptionList(optionList);
    if(err != 0) {
        fprintf(stderr,"PGP Error %d freeing encryption option list\n",err);
    }
    
    return self;
}



/***************************************  DECRYPTION ROUTINES ***********************************/


- (NSData *)decryptData:(NSData *)cipherData
             passPhrase:(NSString *)phraseString
                  error:(int *)errorCode;
{
    PGPOptionListRef optionList;
    char *plainBuffer,errorText[256];
    PGPSize plainLength;
    PGPSize maxPlainLength = 1000000000;
    NSData *plainData;
    PGPError err;

    *errorCode = 0;

    [self signatureStatus:PGP6_NOSIG];
    [self signator:@""];
    [self signatureDate:nil];

    /* Make a new Option List */
    err = PGPNewOptionList(context, &optionList);
    if(err != 0) {
        fprintf(stderr,"PGP Error %d creating Decode option list\n",err);
        *errorCode = err;
        return nil;
    }

    if((phraseString != nil) && (![phraseString isEqual:@""])) {
        err = PGPAppendOptionList(optionList,
                                  PGPOPassphrase(context,[phraseString cString]),
                                  PGPOLastOption(context));

        if(err != 0) {
            fprintf(stderr,"PGP Error %d appending ascii armor option to option list\n",err);
            *errorCode = err;
            return nil;
        }
    }

    err = PGPAppendOptionList(optionList,
                              PGPOInputBuffer(context,[cipherData bytes],[cipherData length]),
                              PGPOAllocatedOutputBuffer(context,
                                                        (void **)&plainBuffer,
                                                        maxPlainLength,
                                                        (PGPSize *)&plainLength),
                              PGPOKeySetRef(context, keyRingSet),
                              PGPOEventHandler(context, decodeEvent, self),
                              PGPOLastOption(context));

    if(err != 0) {
        fprintf(stderr,"PGP Error %d appending all other options to Decode option list\n",err);
        *errorCode = err;
        return nil;
    }

    err = PGPDecode(context, optionList, PGPOLastOption(context));

    if((err != 0)) {
        PGPGetErrorString(err, 256, errorText);
        fprintf(stderr,"PGP error %d decrypting data: %s\n",err,errorText);
        *errorCode = err;
        return nil;
    }

    err = PGPFreeOptionList(optionList);
    if(err != 0) {
        fprintf(stderr,"PGP Error %d freeing decode option list\n",err);
    }
    
    plainData = [NSData dataWithBytes:plainBuffer length:plainLength];

    PGPFreeData(plainBuffer);
    return plainData;
}


- decryptFile:(NSString *)cipherFile passPhrase:(NSString *)phraseString
                                      plainFile:(NSString *)plainFile
                                          error:(int *)errorCode;
{
    PGPOptionListRef optionList;
    char errorText[256];
    PGPFileSpecRef plainFileSpec,cipherFileSpec;
    PGPError err;

    *errorCode = 0;

    [self signatureStatus:PGP6_NOSIG];
    [self signator:@""];
    [self signatureDate:nil];

    /* Make a new Option List */
    err = PGPNewOptionList(context, &optionList);
    if(err != 0) {
        fprintf(stderr,"PGP Error %d creating Decode option list\n",err);
        *errorCode = err;
        return nil;
    }

    if((phraseString != nil) && (![phraseString isEqual:@""])) {
        err = PGPAppendOptionList(optionList,
                                  PGPOPassphrase(context,[phraseString cString]),
                                  PGPOLastOption(context));

        if(err != 0) {
            fprintf(stderr,"PGP Error %d appending ascii armor option to option list\n",err);
            *errorCode = err;
            return nil;
        }
    }

    /* Get the plainText FileSpecRef */
    err = PGPNewFileSpecFromFullPath(context, [plainFile cString], &plainFileSpec);
    if(err != 0) {
        fprintf(stderr,"PGP Error %d creating plain text file spec ref list\n",err);
        *errorCode = err;
        return nil;
    }

    /* Get the cipherText FileSpecRef */
    err = PGPNewFileSpecFromFullPath(context, [cipherFile cString], &cipherFileSpec);
    if(err != 0) {
        fprintf(stderr,"PGP Error %d creating cipher text file spec ref list\n",err);
        *errorCode = err;
        return nil;
    }

    err = PGPAppendOptionList(optionList,
                              PGPOInputFile(context,cipherFileSpec),
                              PGPOOutputFile(context,plainFileSpec),
                              PGPOKeySetRef(context, keyRingSet),
                              PGPOEventHandler(context, decodeEvent, self),
                              PGPOLastOption(context));

    if(err != 0) {
        fprintf(stderr,"PGP Error %d appending all other options to Decode option list\n",err);
        *errorCode = err;
        return nil;
    }

    err = PGPDecode(context, optionList, PGPOLastOption(context));

    if((err != 0)) {
        PGPGetErrorString(err, 256, errorText);
        fprintf(stderr,"PGP error %d decrypting data: %s\n",err,errorText);
        *errorCode = err;
        return nil;
    }

    err = PGPFreeFileSpec(plainFileSpec);
    if(err != 0) {
        fprintf(stderr,"PGP Error %d freeing plain file spec\n",err);
    }

    err = PGPFreeFileSpec(cipherFileSpec);
    if(err != 0) {
        fprintf(stderr,"PGP Error %d freeing cipher file spec\n",err);
    }

    err = PGPFreeOptionList(optionList);
    if(err != 0) {
        fprintf(stderr,"PGP Error %d freeing decode option list\n",err);
    }
    
    return self;
}



/***************************************  SIGNATURE ROUTINES ***********************************/

- signatureStatus:(int)newStatus
{
    signatureStatus = newStatus;
    return self;
}


- signator:(NSString *)newSignator
{
    if(signator) [signator release];
    signator = [newSignator copy];
    return self;
}

- signatureDate:(NSDate *)newSigDate
{
    if(signatureDate) [signatureDate release];
    signatureDate = [newSigDate copy];
    return self;
}


- (int)signatureStatus
{
    return signatureStatus;
}


- (NSString *)signator
{
    return [[signator copy] autorelease];
}

- (NSDate *)signatureDate
{
    return [[signatureDate copy] autorelease];
}




- (NSData *)signData:(NSData *)docData 
		 passPhrase:(NSString *)phraseString 
                      error:(int *)errorCode;
{
    PGPOptionListRef optionList;
    PGPError err;
    PGPKeyRef sigKey;
    char *signatureBuffer;
    NSData *signatureData;
    char errorText[256];
    PGPSize maxSigLength,sigLength;

    *errorCode = 0;
    maxSigLength = 100000000;
    
    /* Make a new Option List */
    err = PGPNewOptionList(context, &optionList);
    if(err != 0) {
        fprintf(stderr,"PGP Error %d creating encryption option list\n",err);
        *errorCode = err;
        return nil;
    }

    /* Append necessary options */
    err = PGPAppendOptionList(optionList,
                              PGPOClearSign(context, TRUE),
                              PGPOArmorOutput(context, TRUE),
                              PGPOEventHandler(context, encodeEvent, self),
                              PGPOInputBuffer(context, (char *)[docData bytes], (int)[docData length]),
                              PGPOAllocatedOutputBuffer(context, (void **)&signatureBuffer,
                                                        maxSigLength,&sigLength),
                              PGPOLastOption(context));

    if(err != 0) {
        fprintf(stderr,"PGP Error %d appending triplet options to option list\n",err);
        *errorCode = err;
        return nil;
    }

    if(clearSign == YES) {
        err = PGPAppendOptionList(optionList,
                                  PGPOClearSign(context, TRUE),
                                  PGPOLastOption(context));

        if(err != 0) {
            fprintf(stderr,"PGP Error %d appending ClearSign to option list\n",err);
            *errorCode = err;
            return nil;
        }
    }
    else if(detachSignature==YES) {
        err = PGPAppendOptionList(optionList,
                                  PGPODetachedSig(context,
                                                  PGPOLastOption(context)),
                                  PGPOLastOption(context));

        if(err != 0) {
            fprintf(stderr,"PGP Error %d appending Detached Sig to option list\n",err);
            *errorCode = err;
            return nil;
        }
    }

    err = PGPGetDefaultPrivateKey(keyRingSet, &sigKey);
    if((err != 0)) {
        PGPGetErrorString(err, 256, errorText);
        fprintf(stderr,"PGP error %d getting default signing key: %s\n",err,errorText);
        *errorCode = err;
        return nil;
    }


    err = PGPAppendOptionList(optionList,
                              PGPOSignWithKey(context,
                                              sigKey,
                                              PGPOPassphrase(context, [phraseString cString]),
                                              PGPOLastOption(context)),
                              PGPOLastOption(context));

    if(err != 0) {
        fprintf(stderr,"PGP Error %d appending signature option to option list\n",err);
        *errorCode = err;
        return nil;
    }

    err = PGPEncode(context,
                    optionList,
                    PGPOLastOption(context));
    
    if(err != 0) {
        fprintf(stderr,"PGP Error %d on encrypt\n",err);
        *errorCode = err;
        return nil;
    }

    err = PGPFreeOptionList(optionList);
    if(err != 0) {
        fprintf(stderr,"PGP Error %d freeing encryption option list\n",err);
    }

    signatureData = [NSData dataWithBytes:signatureBuffer length:sigLength];
    PGPFreeData(signatureBuffer);

    return signatureData;
}
    


- signFile:(NSString *)docFile toFile:(NSString *)sigFile
                           passPhrase:(NSString *)phraseString
                                error:(int *)errorCode;
{
    PGPOptionListRef optionList;
    PGPFileSpecRef docFileSpec,sigFileSpec;
    PGPError err;
    PGPKeyRef sigKey;
    char errorText[256];

    *errorCode = 0;
    
    /* Get the docFile FileSpecRef */
    err = PGPNewFileSpecFromFullPath(context, [docFile cString], &docFileSpec);
    if(err != 0) {
        fprintf(stderr,"PGP Error %d creating document signature file spec ref list\n",err);
        *errorCode = err;
        return nil;
    }

    /* Get the sigFile FileSpecRef */
    err = PGPNewFileSpecFromFullPath(context, [sigFile cString], &sigFileSpec);
    if(err != 0) {
        fprintf(stderr,"PGP Error %d creating signature file spec ref list\n",err);
        *errorCode = err;
        return nil;
    }

    /* Make a new Option List */
    err = PGPNewOptionList(context, &optionList);
    if(err != 0) {
        fprintf(stderr,"PGP Error %d creating encryption option list\n",err);
        *errorCode = err;
        return nil;
    }

    /* Append necessary options */
    err = PGPAppendOptionList(optionList,
                              PGPOArmorOutput(context, TRUE),
                              PGPOEventHandler(context, encodeEvent, self),
                              PGPOInputFile(context,docFileSpec),
                              PGPOOutputFile(context,sigFileSpec),
                              PGPOLastOption(context));

    if(err != 0) {
        fprintf(stderr,"PGP Error %d appending triplet options to option list\n",err);
        *errorCode = err;
        return nil;
    }

    if(clearSign == YES) {
        err = PGPAppendOptionList(optionList,
                                  PGPOClearSign(context, TRUE),
                                  PGPOLastOption(context));

        if(err != 0) {
            fprintf(stderr,"PGP Error %d appending ClearSign to option list\n",err);
            *errorCode = err;
            return nil;
        }
    }
    else if(detachSignature==YES) {
        err = PGPAppendOptionList(optionList,
                                  PGPODetachedSig(context,
                                                  PGPOLastOption(context)),
                                  PGPOLastOption(context));

        if(err != 0) {
            fprintf(stderr,"PGP Error %d appending Detached Sig to option list\n",err);
            *errorCode = err;
            return nil;
        }
    }

    err = PGPGetDefaultPrivateKey(keyRingSet, &sigKey);
    if(err == 0) {
        err = PGPAppendOptionList(optionList,
                                  PGPOSignWithKey(context,
                                                  sigKey,
                                                  PGPOPassphrase(context, [phraseString cString]),
                                                  PGPOLastOption(context)),
                                  PGPOLastOption(context));

        if(err != 0) {
            fprintf(stderr,"PGP Error %d appending signature option to option list\n",err);
            *errorCode = err;
            doSignature = NO;
            clearSign = NO;
        }
    }
    else {
        PGPGetErrorString(err, 256, errorText);
        fprintf(stderr,"PGP error %d getting default signing key: %s\n",err,errorText);
        *errorCode = err;
    }

    err = PGPEncode(context,
                    optionList,
                    PGPOLastOption(context));

    if(err != 0) {
        fprintf(stderr,"PGP Error %d on encrypt\n",err);
        *errorCode = err;
        return nil;
    }

    err = PGPFreeFileSpec(docFileSpec);
    if(err != 0) {
        fprintf(stderr,"PGP Error %d freeing document file spec\n",err);
    }

    err = PGPFreeFileSpec(sigFileSpec);
    if(err != 0) {
        fprintf(stderr,"PGP Error %d freeing signature file spec\n",err);
    }

    err = PGPFreeOptionList(optionList);
    if(err != 0) {
        fprintf(stderr,"PGP Error %d freeing encryption option list\n",err);
    }

    return self;
}



- (NSData *)verifySignatureData:(NSData *)docData
                     passPhrase:(NSString *)phraseString
                          error:(int *)errorCode;
{
    *errorCode = 0;
    return [self decryptData:docData passPhrase:nil error:errorCode];
}


- (NSData *)verifySignatureFile:(NSString *)docFile
                     passPhrase:(NSString *)phraseString
                          error:(int *)errorCode;
{
    NSString *outFile = [NSString stringWithString:@"/dev/null"];

    *errorCode = 0;
    return [self decryptFile:docFile passPhrase:phraseString plainFile:outFile error:errorCode];
}


- verifyDetachedSignatureData:(NSData *)sigData
	          againstData:(NSData *)docData
                   passPhrase:(NSString *)phraseString
	                error:(int *)errorCode
{
    PGPError err;
    char *plainBuffer;
    PGPSize maxPlainLength = 1000000000;
    PGPSize plainLength;
    char errorText[256];

    *errorCode = 0;

    err = PGPDecode(context,
                    PGPOInputBuffer(context,(char *)[sigData bytes],(int)[sigData length]),
                    PGPOAllocatedOutputBuffer(context,
                                              (void **)&plainBuffer,
                                              maxPlainLength,
                                              (PGPSize *)&plainLength),
                    PGPOKeySetRef(context, keyRingSet),
                    PGPOPassphrase(context, [phraseString cString]),
                    PGPODetachedSig(context,
                                    PGPOInputBuffer(context,
                                                    (char *)[docData bytes], (int)[docData length]),
                                    PGPOLastOption(context)),
                    PGPOEventHandler(context, decodeEvent, self),
                    PGPOLastOption(context));

    if((err != 0)) {
        PGPGetErrorString(err, 256, errorText);
        fprintf(stderr,"PGP error %d verifying detached signature: %s\n",err,errorText);
        *errorCode = err;
        return nil;
    }

    PGPFreeData(plainBuffer);
    return self;
}

	
- verifyDetachedSignatureFile:(NSString *)sigFile
		  againstFile:(NSString *)docFile
                   passPhrase:(NSString *)phraseString
			error:(int *)errorCode
{
    PGPError err;
    PGPFileSpecRef sigFileSpec,docFileSpec;
    char *plainBuffer;
    PGPSize maxPlainLength = 1000000000;
    PGPSize plainLength;
    char errorText[256];

    *errorCode = 0;

    err = PGPNewFileSpecFromFullPath(context, [sigFile cString], &sigFileSpec);
    if(err != 0) {
        fprintf(stderr,"PGP Error %d creating sig file spec ref\n",err);
        *errorCode = err;
        return nil;
    }

    err = PGPNewFileSpecFromFullPath(context, [docFile cString], &docFileSpec);
    if(err != 0) {
        fprintf(stderr,"PGP Error %d creating document file spec ref\n",err);
        *errorCode = err;
        return nil;
    }

    err = PGPDecode(context,
                    PGPOInputFile(context,sigFileSpec),
                    PGPOAllocatedOutputBuffer(context,
                                              (void **)&plainBuffer,
                                              maxPlainLength,
                                              (PGPSize *)&plainLength),
                    PGPOKeySetRef(context, keyRingSet),
                    PGPOPassphrase(context, [phraseString cString]),
                    PGPODetachedSig(context,
                                    PGPOInputFile(context,docFileSpec),
                                    PGPOLastOption(context)),
                    PGPOEventHandler(context, decodeEvent, self),
                    PGPOLastOption(context));

    if((err != 0)) {
        PGPGetErrorString(err, 256, errorText);
        fprintf(stderr,"PGP error %d verifying detached signature: %s\n",err,errorText);
        *errorCode = err;
        return nil;
    }

    PGPFreeFileSpec(sigFileSpec);
    PGPFreeFileSpec(docFileSpec);
    PGPFreeData(plainBuffer);
    return self;
}




/***************************************  KEY MANAGEMENT ROUTINES ***********************************/


- (int)userIDExists:(NSString *)userID error:(int *)errorCode
{
    PGPFilterRef idFilter;
    PGPKeySetRef tmpKeySet;
    char errorText[256];
    int numKeys;
    PGPError err;

    *errorCode = 0;
    
    err = PGPNewUserIDStringFilter(context, [userID cString], kPGPMatchSubString, &idFilter);
    if((err != 0)) {
        fprintf(stderr,"PGP error %d creating userID filter\n",err);
        *errorCode = err;
        return NO;
    }

    /* Does this return an error if recipient isn't on keyring? */
    err = PGPFilterKeySet(keyRingSet, idFilter, &tmpKeySet);
    if((err != 0)) {
        PGPGetErrorString(err, 256, errorText);
        fprintf(stderr,"PGP error %d Filtering KeySet: %s\n",err,errorText);
    }

    /* Check whether you found the key for this user */
    PGPCountKeys(tmpKeySet, &numKeys);

    PGPFreeFilter(idFilter);
    PGPFreeKeySet(tmpKeySet);

    return numKeys;
}


    

- (NSMutableArray *)getAllUserIDs:(int *)errorCode
{
    NSMutableArray *userIDArray;
    PGPKeyListRef keyList;
    PGPKeyIterRef keyIter;
    PGPKeyRef currentKey;
    char errorText[256];
    PGPError err;

    *errorCode = 0;
    
    /* Create a new array */
    userIDArray = [[NSMutableArray alloc] init];

    /* Create a key list */
    err = PGPOrderKeySet(keyRingSet, kPGPAnyOrdering, &keyList);
    if((err != 0)) {
        PGPGetErrorString(err, 256, errorText);
        fprintf(stderr,"PGP error %d creating key list: %s\n",err,errorText);
        *errorCode = err;
        return nil;
    }

    /* Create a new key iterator */
    err = PGPNewKeyIter(keyList,&keyIter);
    if(err != 0) {
        PGPGetErrorString(err, 256, errorText);
        fprintf(stderr,"PGP Error %d creating new key iterator for publicKey UserIDs: %s\n",
                err,errorText);
        return nil;
    }

    /* Iterate through keyring and collect userID's */
    while((err = PGPKeyIterNext(keyIter, &currentKey)) != kPGPError_EndOfIteration) {
        char userName[128];
        PGPSize nameLength;

        if(err != 0) {
            PGPGetErrorString(err, 256, errorText);
            fprintf(stderr,"PGP Error %d iterating through public key: %s\n",err,errorText);
            return nil;
        }

        err = PGPGetPrimaryUserIDNameBuffer(currentKey, (PGPSize)128, userName, &nameLength);
        if(err != 0) {
            PGPGetErrorString(err, 256, errorText);
            fprintf(stderr,"PGP Error %d getting userID from key: %s\n",err,errorText);
            continue;
        }

        [userIDArray addObject:[NSString stringWithCString:userName]];
    }

    PGPFreeKeyIter(keyIter);
    PGPFreeKeyList(keyList);

    return [userIDArray autorelease];
}


- (NSMutableArray *)getUserIDsMatching:(NSString *)userID error:(int *)errorCode
{
    PGPFilterRef idFilter;
    PGPKeySetRef tmpKeySet;
    PGPKeyListRef keyList;
    PGPKeyIterRef keyIter;
    NSMutableArray *userIDArray;
    PGPKeyRef currentKey;
    char errorText[256];
    PGPError err;

    *errorCode = 0;
    
    err = PGPNewUserIDStringFilter(context, [userID cString], kPGPMatchSubString, &idFilter);
    if((err != 0)) {
        fprintf(stderr,"PGP error %d creating userID filter\n",err);
        *errorCode = err;
        return NO;
    }

    /* Does this return an error if recipient isn't on keyring? */
    err = PGPFilterKeySet(keyRingSet, idFilter, &tmpKeySet);
    if((err != 0)) {
        PGPGetErrorString(err, 256, errorText);
        fprintf(stderr,"PGP error %d Filtering KeySet: %s\n",err,errorText);
    }

    /* Create a key list */
    err = PGPOrderKeySet(tmpKeySet, kPGPAnyOrdering, &keyList);
    if((err != 0)) {
        PGPGetErrorString(err, 256, errorText);
        fprintf(stderr,"PGP error %d creating key list: %s\n",err,errorText);
        *errorCode = err;
        return nil;
    }

    /* Create a new key iterator */
    err = PGPNewKeyIter(keyList,&keyIter);
    if(err != 0) {
        PGPGetErrorString(err, 256, errorText);
        fprintf(stderr,"PGP Error %d creating new key iterator for publicKey UserIDs: %s\n",
                err,errorText);
        return nil;
    }

    /* Create a new array */
    userIDArray = [[NSMutableArray alloc] init];

    /* Iterate through keyring and collect userID's */
    while((err = PGPKeyIterNext(keyIter, &currentKey)) != kPGPError_EndOfIteration) {
        char userName[128];
        PGPSize nameLength;

        if(err != 0) {
            PGPGetErrorString(err, 256, errorText);
            fprintf(stderr,"PGP Error %d iterating through public key: %s\n",err,errorText);
            return nil;
        }

        err = PGPGetPrimaryUserIDNameBuffer(currentKey, (PGPSize)128, userName, &nameLength);
        if(err != 0) {
            PGPGetErrorString(err, 256, errorText);
            fprintf(stderr,"PGP Error %d getting userID from key: %s\n",err,errorText);
            continue;
        }

        [userIDArray addObject:[NSString stringWithCString:userName]];
    }

    PGPFreeKeyIter(keyIter);
    PGPFreeKeyList(keyList);
    PGPFreeKeySet(tmpKeySet);
    PGPFreeFilter(idFilter);

    return [userIDArray autorelease];
}    



- (NSString *)fingerprintForUserID:(NSString *)userID error:(int *)errorCode
{
    NSString *fpString;
    PGPFilterRef userIDFilter;
    PGPKeySetRef userKeySet;
    PGPKeyListRef userKeyList;
    PGPKeyIterRef userKeyIter;
    PGPKeyRef userKey;
    unsigned char fp[20];
    char errorText[256];
    PGPSize fpsize;
    PGPError err;

    *errorCode = 0;
    
    err = PGPNewUserIDStringFilter(context, [userID cString], kPGPMatchSubString, &userIDFilter);
    if((err != 0)) {
        PGPGetErrorString(err,256,errorText);
        fprintf(stderr,"PGP error %d in creating userID filter for key fingerprint\n",err);
        *errorCode = err;
        return nil;
    }

    err = PGPFilterKeySet(keyRingSet, userIDFilter, &userKeySet);
    if((err != 0)) {
        PGPGetErrorString(err, 256, errorText);
        fprintf(stderr,"PGP error %d Filtering KeySet: %s\n",err,errorText);
        *errorCode = err;
        return nil;
    }

    err = PGPOrderKeySet(userKeySet, kPGPAnyOrdering, &userKeyList);
    if((err != 0)) {
        PGPGetErrorString(err, 256, errorText);
        fprintf(stderr,"PGP error %d Ordering KeySet for fingerprint: %s\n",err,errorText);
        *errorCode = err;
        return nil;
    }

    err = PGPNewKeyIter(userKeyList, &userKeyIter);
    if((err != 0)) {
        PGPGetErrorString(err, 256, errorText);
        fprintf(stderr,"PGP error %d creating new Key iterator for fingerprint: %s\n",err,errorText);
        *errorCode = err;
        return nil;
    }

    err = PGPKeyIterNext(userKeyIter, &userKey);
    if((err != 0)) {
        PGPGetErrorString(err, 256, errorText);
        fprintf(stderr,"PGP error %d getting next key from list for fingerprint: %s\n",err,errorText);
        *errorCode = err;
        return nil;
    }

    err = PGPGetKeyPropertyBuffer(userKey, kPGPKeyPropFingerprint, 20, fp, &fpsize);
    if((err != 0)) {
        PGPGetErrorString(err, 256, errorText);
        fprintf(stderr,"PGP error %d extracting fingerprint from key: %s\n",err,errorText);
        *errorCode = err;
        return nil;
    }


    fpString = [NSString stringWithFormat:@"%02x%02x %02x%02x %02x%02x %02x%02x %02x%02x %02x%02x %02x%02x %02x%02x %02x%02x %02x%02x", fp[0],fp[1],fp[2],fp[3],fp[4],fp[5],fp[6],fp[7],fp[8],fp[9],fp[10],fp[11],fp[12],fp[13],fp[14],fp[15],fp[16],fp[17],fp[18],fp[19]];

    PGPFreeKeyIter(userKeyIter);
    PGPFreeKeyList(userKeyList);
    PGPFreeKeySet(userKeySet);
    PGPFreeFilter(userIDFilter);

    return fpString;
}
    


- importKeyData:(NSData *)keyData error:(int *)errorCode
{
    PGPKeySetRef newKeySet;
    BOOL writeFlag = YES;
    char errorText[256];
    PGPError err;

    *errorCode = 0;
    
    /* Need to open keyrings with write access */
    [self openKeyringFiles:writeFlag];

    /* Import key data into a keyset */
    err = PGPImportKeySet(context,
                          &newKeySet,
                          PGPOInputBuffer(context,
                                          (char *)[keyData bytes], (int)[keyData length]),
                          PGPOLastOption(context));

    if(err != 0) {
        PGPGetErrorString(err, 256, errorText);
        fprintf(stderr,"PGP Error %d importing key data into new keyset: %s\n",err,errorText);
        *errorCode = err;
        return nil;
    }

    /* Add new key to keyring */
    err = PGPAddKeys(newKeySet, keyRingSet);
    if(err != 0) {
        PGPGetErrorString(err, 256, errorText);
        fprintf(stderr,"PGP Error %d adding keys to keyRing: %s\n",err,errorText);
        *errorCode = err;
        return nil;
    }

    /* Free the new keyset */
    PGPFreeKeySet(newKeySet);

    /* Commit changes to disk */
    err = PGPCommitKeyRingChanges(keyRingSet);
    if(err != 0) {
        PGPGetErrorString(err, 256, errorText);
        fprintf(stderr,"PGP Error %d updating keyrings on disk: %s\n",err,errorText);
        *errorCode = err;
    }

    /* Open keyrings back up without write access */
    writeFlag = NO;
    [self openKeyringFiles:writeFlag];

    return self;
}


- importKeyFile:(NSString *)keyFile error:(int *)errorCode
{
    NSData *keyData = [NSData dataWithContentsOfFile:keyFile];
    *errorCode = 0;
    [self importKeyData:keyData error:errorCode];
    return self;
}


- (NSData *)exportKeyData:(NSString *)userID error:(int *)errorCode
{
    PGPKeySetRef newKeySet;
    PGPFilterRef userFilter;
    NSData *keyData;
    char errorText[256];
    char *exportBuffer;
    PGPSize maxExportLength,exportLength;
    int numKeys;
    PGPError err;

    *errorCode = 0;
    maxExportLength = 1000000000;

    /* Get the key for the userID */
    err = PGPNewUserIDStringFilter(context,
                                   [userID cString], kPGPMatchSubString, &userFilter);
    if((err != 0)) {
        PGPGetErrorString(err, 256, errorText);
        fprintf(stderr,"PGP error %d creating userID filter: %s\n",err,errorText);
        *errorCode = err;
        return nil;
    }

    /* Does this return an error if recipient isn't no keyring? */
    err = PGPFilterKeySet(keyRingSet, userFilter, &newKeySet);
    if((err != 0)) {
        PGPGetErrorString(err, 256, errorText);
        fprintf(stderr,"PGP error %d Filtering KeySet: %s\n",err,errorText);
        return nil;
    }

    /* Check whether you found the key for this user */
    PGPCountKeys(newKeySet, &numKeys);
    if(numKeys == 0) {
        fprintf(stderr,"Can't find user |%s| on keyring\n",[userID cString]);
        return nil;
    }

    err = PGPFreeFilter(userFilter);
    if(err != 0) {
        PGPGetErrorString(err, 256, errorText);
        fprintf(stderr,"PGP Error %d freeing key filter for user %s: %s\n",
                err,[userID cString],errorText);
    }

    err = PGPExportKeySet(newKeySet,
                          PGPOAllocatedOutputBuffer(context, (void **)&exportBuffer,
                                                    maxExportLength,(PGPSize *)&exportLength),
                          PGPOLastOption(context));

    if(err != 0) {
        PGPGetErrorString(err, 256, errorText);
        fprintf(stderr,"PGP Error %d exporting key data for user %s: %s\n",
                err,[userID cString],errorText);
    }

    keyData = [NSData dataWithBytes:exportBuffer length:exportLength];

    PGPFreeKeySet(newKeySet);
    PGPFreeData(exportBuffer);

    return keyData;
}


- removeKeyForUserID:(NSString *)userID error:(int *)errorCode
{
    PGPKeySetRef newKeySet;
    PGPFilterRef userFilter;
    char errorText[256];
    BOOL writeFlag = YES;
    PGPError err;

    *errorCode = 0;
    
    /* Create new userID filter */
    err = PGPNewUserIDStringFilter(context,
                                   [userID cString], kPGPMatchSubString, &userFilter);
    if((err != 0)) {
        PGPGetErrorString(err, 256, errorText);
        fprintf(stderr,"PGP error %d creating userID filter: %s\n",err,errorText);
        *errorCode = err;
        return nil;
    }

    /* Get the key for the userID */
    err = PGPFilterKeySet(keyRingSet, userFilter, &newKeySet);
    if((err != 0)) {
        PGPGetErrorString(err, 256, errorText);
        fprintf(stderr,"PGP error %d Filtering KeySet: %s\n",err,errorText);
        return nil;
    }

    err = PGPFreeFilter(userFilter);
    if(err != 0) {
        PGPGetErrorString(err, 256, errorText);
        fprintf(stderr,"PGP Error %d freeing key filter for user %s: %s\n",
                err,[userID cString],errorText);
    }

    /* Open keyRing with write access */
    [self openKeyringFiles:writeFlag];

    /* Remove the keys */
    err = PGPRemoveKeys(newKeySet, keyRingSet);
    if(err == 0) {
        /* Commit changes to disk */
        err = PGPCommitKeyRingChanges(keyRingSet);
        if(err != 0) {
            PGPGetErrorString(err, 256, errorText);
            fprintf(stderr,"PGP Error %d updating keyrings on disk: %s\n",err,errorText);
            *errorCode = err;
        }
    }
    else {
        PGPGetErrorString(err, 256, errorText);
        fprintf(stderr,"PGP Error %d removing key for user %s: %s\n",
                err,[userID cString],errorText);
    }
        
    /* Free the new keyset */
    PGPFreeKeySet(newKeySet);

    /* Open keyrings back up without write access */
    writeFlag = NO;
    [self openKeyringFiles:writeFlag];

    return self;
}


/* This method examines keyData and extracts the userID
   and fingerprint from it.  The userID and fingerprint
   can later be queried using the methods above
*/
- assessKeyFromData:(NSData *)keyData error:(int *)errorCode
{
    PGPError err;
    char *plainBuffer;
    PGPSize maxPlainLength = 1000000000;
    PGPSize plainLength;
    char errorText[256];

    *errorCode = 0;

    err = PGPDecode(context,
                    PGPOInputBuffer(context,(char *)[keyData bytes],(int)[keyData length]),
                    PGPOAllocatedOutputBuffer(context,
                                              (void **)&plainBuffer,
                                              maxPlainLength,
                                              (PGPSize *)&plainLength),
                    PGPOKeySetRef(context, keyRingSet),
                    PGPOEventHandler(context, decodeEvent, self),
                    PGPOSendEventIfKeyFound(context, TRUE),
                    PGPOLastOption(context));

    if((err != 0)) {
        PGPGetErrorString(err, 256, errorText);
        fprintf(stderr,"PGP error %d verifying detached signature: %s\n",err,errorText);
        *errorCode = err;
        return nil;
    }

    PGPFreeData(plainBuffer);
    return self;
}    


- setNewKeyUserID:(NSString *)newUserID
{
    [newUserID retain];
    if(newKeyUserID) [newKeyUserID release];
    newKeyUserID = newUserID;
    return self;
}


- setNewKeyFingerprint:(NSString *)newFingerprint
{
    [newFingerprint retain];
    if(newKeyFingerprint) [newKeyFingerprint release];
    newKeyFingerprint = newFingerprint;
    return self;
}

- (NSString *)newKeyUserID
{
    return newKeyUserID;
}


- (NSString *)newKeyFingerprint
{
    return newKeyFingerprint;
}


- setDefaultKey:(NSString *)defaultUserID error:(int *)errorCode
{
    PGPFilterRef newFilter;
    PGPKeySetRef newKeySet;
    PGPKeyListRef newKeyList;
    PGPKeyIterRef newKeyIter;
    PGPKeyRef defaultKey;
    char errorText[256];
    PGPError err;

    *errorCode = 0;
    
    /* First, create a filter for signorID */
    err = PGPNewUserIDStringFilter(context,
                                   [defaultUserID cString], kPGPMatchSubString, &newFilter);
    if((err != 0)) {
        PGPGetErrorString(err, 256, errorText);
        fprintf(stderr,"PGP error %d creating default userID filter: %s\n",err,errorText);
        *errorCode = err;
        return nil;
    }

    /* Get keys for signorID */
    err = PGPFilterKeySet(keyRingSet, newFilter, &newKeySet);
    if((err != 0)) {
        PGPGetErrorString(err, 256, errorText);
        fprintf(stderr,"PGP error %d Filtering KeySet: %s\n",err,errorText);
        return nil;
    }

    err = PGPFreeFilter(newFilter);
    if(err != 0) {
        PGPGetErrorString(err, 256, errorText);
        fprintf(stderr,"PGP Error %d freeing key filter for user %s: %s\n",
                err,[defaultUserID cString],errorText);
    }

    /* Pick the first key from the filter */
    err = PGPOrderKeySet(newKeySet, kPGPAnyOrdering, &newKeyList);
    if((err != 0)) {
        PGPGetErrorString(err, 256, errorText);
        fprintf(stderr,"PGP error %d Ordering KeySet for setting default key: %s\n",err,errorText);
        *errorCode = err;
        return nil;
    }

    err = PGPNewKeyIter(newKeyList, &newKeyIter);
    if((err != 0)) {
        PGPGetErrorString(err, 256, errorText);
        fprintf(stderr,"PGP error %d creating new Key iterator for signing key: %s\n",err,errorText);
        *errorCode = err;
        return nil;
    }

    err = PGPKeyIterNext(newKeyIter, &defaultKey);
    if((err != 0)) {
        PGPGetErrorString(err, 256, errorText);
        fprintf(stderr,"PGP error %d getting next key from list for default key: %s\n",err,errorText);
        *errorCode = err;
        return nil;
    }

    /* Make this the default key */
    err = PGPSetDefaultPrivateKey(defaultKey);
    if(err == 0) {
        err = PGPsdkSavePrefs(context);
        if((err != 0)) {
            PGPGetErrorString(err, 256, errorText);
            fprintf(stderr,"PGP error %d Saving preferences: %s\n",err,errorText);
            *errorCode = err;
        }
    }
    else {
        PGPGetErrorString(err, 256, errorText);
        fprintf(stderr,"PGP error %d getting next key from list for default key: %s\n",err,errorText);
        *errorCode = err;
        return nil;
    }

    err = PGPFreeKeyIter(newKeyIter);
    err = PGPFreeKeyList(newKeyList);
    err = PGPFreeKeySet(newKeySet);

    return self;
}


- (NSData *)generateNewKeyPair:(NSString *)userID
                       keySize:(PGPUInt32)keySize
                       entropy:(int *)entropyPool
                   entropySize:(int)entropySize
                    passPhrase:(NSString *)passphrase
                         error:(int *)errorCode
{
    int ientropy;
    PGPUInt32 requiredEntropy;
    PGPKeyRef newKey;
    PGPSubKeyRef newSubkey;
    PGPKeySetRef newKeySet;
    char errorText[256];
    PGPSize exportLength,maxExportLength;
    char *exportBuffer;
    NSData *keyData;
    BOOL writeFlag;
    PGPError err;
    PGPError keyGenEvent(PGPContextRef, PGPEvent *, PGPUserValue);

    *errorCode = 0;
    maxExportLength = 1000000000;
    if(entropyLength > 0) free(entropyBucket);
    entropyBucket = (int *)malloc(entropySize*sizeof(int));
    entropyLength = entropySize;

    for(ientropy=0; ientropy<entropySize; ientropy++)
        entropyBucket[ientropy] = entropyPool[ientropy];

    requiredEntropy = PGPGetKeyEntropyNeeded(context,
                                             PGPOKeyGenParams(context, kPGPPublicKeyAlgorithm_DSA, keySize),
                                             PGPOLastOption(context));

    requiredEntropy *= 2;

    do {
        if((entropyBucket == NULL) || (entropyLength == 0)) {
            usleep(random() & 0x0000FFFF);
        }
        else {
            if((ientropy >= entropyLength) || (entropyBucket[ientropy] == 0)) {
                usleep(random() & 0x0000FFFF);
                ientropy++;
            }
            else {
                usleep(entropyBucket[ientropy++] & 0x0000FFFF);
            }
        }
        PGPGlobalRandomPoolAddKeystroke(32 + (random() % 96));

    } while(PGPGlobalRandomPoolGetEntropy() < requiredEntropy);

    writeFlag = YES;
    [self openKeyringFiles:writeFlag];

    fprintf(stderr,"\nGenerating Master key\n");

    err = PGPGenerateKey(context,
                         &newKey,
                         PGPOKeySetRef(context, keyRingSet),
                         PGPOKeyGenParams(context, kPGPPublicKeyAlgorithm_DSA, 1024),
                         PGPOKeyGenName(context, (char *)[userID cString], (PGPSize)[userID length]),
                         PGPOPassphrase(context, [passphrase cString]),
                         PGPOEventHandler(context, keyGenEvent, self),
                         PGPOLastOption(context));

    if(err != 0) {
        PGPGetErrorString(err, 256, errorText);
        fprintf(stderr,"PGP Error %d while generating new master key: %s\n",err,errorText);
        return nil;
    }

    fprintf(stderr,"\nGenerating subkey\n");

    err = PGPGenerateSubKey(context,
                            &newSubkey,
                            PGPOKeyGenMasterKey(context, newKey),
                            PGPOKeyGenParams(context, kPGPPublicKeyAlgorithm_ElGamal, keySize),
                            PGPOPassphrase(context, [passphrase cString]),
                            PGPOEventHandler(context, keyGenEvent, self),
                            PGPOLastOption(context));

    fprintf(stderr,"Done.\n");

    if(err != 0) {
        PGPGetErrorString(err, 256, errorText);
        fprintf(stderr,"PGP Error %d while generating new sub-key: %s\n",err,errorText);
        return nil;
    }

    /* Add this to a keySet, and export it */
    err = PGPNewSingletonKeySet(newKey, &newKeySet);
    if(err != 0) {
        PGPGetErrorString(err, 256, errorText);
        fprintf(stderr,"PGP Error %d creating a singleton keyset: %s\n",err,errorText);
        return nil;
    }

    /* Extract key data from it */
    err = PGPExportKeySet(newKeySet,
                          PGPOAllocatedOutputBuffer(context, (void **)&exportBuffer,
                                                    maxExportLength,(PGPSize *)&exportLength),
                          PGPOLastOption(context));

    if(err != 0) {
        PGPGetErrorString(err, 256, errorText);
        fprintf(stderr,"PGP Error %d exporting key data for user %s: %s\n",
                err,[userID cString],errorText);
    }

    keyData = [NSData dataWithBytes:exportBuffer length:exportLength];

    /* Commit changes to disk */
    err = PGPCommitKeyRingChanges(keyRingSet);
    if(err != 0) {
        PGPGetErrorString(err, 256, errorText);
        fprintf(stderr,"PGP Error %d commiting new key to disk: %s\n",err,errorText);
        *errorCode = err;
    }

    writeFlag = NO;
    [self openKeyringFiles:writeFlag];
    
    return keyData;
}


PGPError decodeEvent(PGPContextRef pgpContext, PGPEvent *triggerEvent, PGPUserValue userData)
{
    PGPError err;
    PGPKeySetRef keySet;
    PGPKeyRef userKey = NULL;
    PGPKeyListRef keyList = NULL;
    PGPKeyIterRef keyIter = NULL;
    char errorText[256];
    char userName[128];
    unsigned char fp[20];
    PGPSize nameLength,fpsize;
    int nkeys;
    NSString *fingerprint,*userID;    
    
    switch (triggerEvent->type) {
      case kPGPEvent_KeyFoundEvent:
          keySet = triggerEvent->data.keyFoundData.keySet;
          err = PGPCountKeys(keySet, &nkeys);

          /* Get a key List */
          err = PGPOrderKeySet(keySet, kPGPAnyOrdering, &keyList);
          if(err != 0) {
              PGPGetErrorString(err, 256, errorText);
              fprintf(stderr,"PGP error %d Ordering KeySet for new key: %s\n",err,errorText);
              return err;
          }

          err = PGPNewKeyIter(keyList, &keyIter);
          if((err != 0)) {
              PGPGetErrorString(err, 256, errorText);
              fprintf(stderr,"PGP error %d creating new Key iterator for new key: %s\n",err,errorText);
              return err;
          }

          err = PGPKeyIterNext(keyIter, &userKey);
          if((err != 0)) {
              PGPGetErrorString(err, 256, errorText);
              fprintf(stderr,"PGP error %d getting next key from list for new key: %s\n",err,errorText);
              return err;
          }

          err = PGPGetPrimaryUserIDNameBuffer(userKey, (PGPSize)128, userName, &nameLength);
          if(err != 0) {
              PGPGetErrorString(err, 256, errorText);
              return err;
          }

          userID = [NSString stringWithCString:userName];
          [(PGPserver *)userData setNewKeyUserID:userID];

          /* Get fingerprint */
          err = PGPGetKeyPropertyBuffer(userKey, kPGPKeyPropFingerprint, 20, fp, &fpsize);
          if((err != 0)) {
              PGPGetErrorString(err, 256, errorText);
              fprintf(stderr,"PGP error %d extracting fingerprint from new key: %s\n",err,errorText);
              return err;
          }

              fingerprint = [NSString stringWithFormat:@"%02x%02x %02x%02x %02x%02x %02x%02x %02x%02x %02x%02x %02x%02x %02x%02x %02x%02x %02x%02x",
              fp[0],fp[1],fp[2],fp[3],fp[4],fp[5],fp[6],fp[7],fp[8],fp[9],fp[10],fp[11],fp[12],fp[13],fp[14],fp[15],fp[16],fp[17],fp[18],fp[19]];

          [(PGPserver *)userData setNewKeyFingerprint:fingerprint];

          PGPFreeKeyIter(keyIter);
          PGPFreeKeyList(keyList);

          break;
          
      case kPGPEvent_ErrorEvent:
	err = triggerEvent->data.errorData.error;
        PGPGetErrorString(err, 256, errorText);
	fprintf(stderr,"Got a PGP Error %d during decode: %s\n",err,errorText);
        return err;
          
      case kPGPEvent_DetachedSignatureEvent:
          /* Prompt user with an open panel for the corresponding doc file */
        break;


    /* Set signature status here */
    case kPGPEvent_SignatureEvent:
        if(triggerEvent->data.signatureData.checked == 1) {
            if(triggerEvent->data.signatureData.verified) {
                char userName[128];
                PGPSize nameLength;
                PGPKeyRef keyID = triggerEvent->data.signatureData.signingKey;
                time_t sigTime = PGPGetStdTimeFromPGPTime(triggerEvent->data.signatureData.creationTime);
                PGPGetPrimaryUserIDNameBuffer(keyID,(PGPSize)128,userName,&nameLength);
                [(PGPserver *)userData signatureStatus:PGP6_GOODSIG];
                [(PGPserver *)userData signator:[NSString stringWithCString:userName]];
                [(PGPserver *)userData signatureDate:[NSDate dateWithTimeIntervalSince1970:sigTime]];
            }
            else {
                [(PGPserver *)userData signatureStatus:PGP6_BADSIG];
                [(PGPserver *)userData signator:@""];
                [(PGPserver *)userData signatureDate:nil];
            }
        }
        else {
            [(PGPserver *)userData signatureStatus:PGP6_SIGNOKEY];
            [(PGPserver *)userData signator:@""];
            [(PGPserver *)userData signatureDate:nil];
        }
            
        break;

      case kPGPEvent_PassphraseEvent:
          return kPGPError_BadPassphrase;

      default:
        break;
    }

    return 0;
}



static PGPError encodeEvent(PGPContextRef pgpContext, PGPEvent *triggerEvent, PGPUserValue userData)
{
    switch(triggerEvent->type) {
      case kPGPEvent_EntropyEvent:
          while(!PGPGlobalRandomPoolHasMinimumEntropy()) {
              usleep(random() & 0x0000FFFF);
              PGPGlobalRandomPoolAddKeystroke(32 + (random() % 96));
          }
        break;

      default:
        break;
    }
    return 0;
}


PGPError keyGenEvent(PGPContextRef pgpContext, PGPEvent *triggerEvent, PGPUserValue userData)
{
    if(triggerEvent->type == kPGPEvent_KeyGenEvent) {
        fprintf(stderr,"%c",triggerEvent->data.keyGenData.state);
        fflush(stderr);
    }
        
    return 1;
}


@end
					
	
